IMPLEMENTATION MODULE Files;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT FIO, FIOStd, Str;
IMPORT Lib2, MD5, UFTP;

TYPE
  TIdx = [ 0 .. cnt ];
  TVldIdx = TIdx[ 1 .. cnt ];

  TFile = POINTER TO TFileDesc;
  TFileDesc = RECORD
    i: TIdx;             (* 0 <==> free *)
    f: FIO.File;
    r: UFTP.TResName;
    l: UFTP.TFgmLen;
    s: UFTP.TFileLen;
    n: UFTP.TFgmIdx;     (* fgm cnt *)
    d: MD5.TDgstDesc;
    p: UFTP.TFileLen;
  END;

CONST
  fileName = 'UFTPSrv.exe';

VAR
  arr: ARRAY TVldIdx OF TFileDesc;
  md5Null: MD5.TDgstDesc;


PROCEDURE Alloc(): TFile;
VAR
  i: TVldIdx;
  f: TFile;
BEGIN
  FOR i := MIN( TVldIdx ) TO MAX( TVldIdx ) DO
    f := ADR( arr[i] );
    IF f^.i = 0 THEN f^.i := i; RETURN f END;
  END;
  RETURN NIL;
END Alloc;

PROCEDURE CalcMD5( f: FIO.File; VAR d: MD5.TDgstDesc ): UFTP.TFileLen;
CONST
  len = 4 * 1024;
VAR
  b: ARRAY [1..len] OF BYTE;
  l: CARDINAL;
  c: MD5.TCtxDesc;
  p: UFTP.TFileLen;
BEGIN
  p := 0;
  MD5.Init( c );
  l := FIO.RdBin( f, b, len );
  WHILE l > 0 DO
    INC( p, UFTP.TFileLen( l ) );
    MD5.Update( c, b, l );
    l := FIO.RdBin( f, b, len );
  END;
  MD5.Final( c, d );
  IF p # FIO.GetPos( f ) THEN (* happens after last successful read (l<len) *) END;
  FIO.Seek( f, 0 );
  RETURN p;
END CalcMD5;


PROCEDURE Open( r-: UFTP.TResName; l: UFTP.TFgmLen ): TFile;
VAR
  f: TFile;
BEGIN
  IF ( r[0] = 0C ) OR ( l = 0 ) THEN RETURN NIL END;
  IF ~ FIO.Exists( r ) THEN RETURN NIL END;
  f := Alloc();
  IF f = NIL THEN RETURN NIL END;
  f^.f := FIO.OpenRead( r );
  Str.Copy( f^.r, r );
  f^.l := l;
  f^.s := CalcMD5( f^.f, f^.d );
  f^.n := f^.s DIV l + ORD( f^.s MOD l > 0 );
  RETURN f;
END Open;

PROCEDURE GetName( f: TFile; VAR r: UFTP.TResName );
BEGIN
  r := '';
  IF f = NIL THEN RETURN END;
  Str.Copy( r, f^.r );
END GetName;

PROCEDURE GetFgl( f: TFile; VAR l: UFTP.TFgmLen );
BEGIN
  l := 0;
  IF f = NIL THEN RETURN END;
  l := f^.l;
END GetFgl;

PROCEDURE GetLen( f: TFile; VAR l: UFTP.TFileLen );
BEGIN
  l := 0;
  IF f = NIL THEN RETURN END;
  l := f^.s;
END GetLen;

PROCEDURE GetMD5( f: TFile; VAR d: MD5.TDgstDesc );
BEGIN
  d := md5Null;
  IF f = NIL THEN RETURN END;
  d := f^.d;
END GetMD5;

PROCEDURE GetPos( f: TFile; VAR p: UFTP.TFileLen );
BEGIN
  p := 0;
  IF f = NIL THEN RETURN END;
  p := f^.p;
END GetPos;

PROCEDURE GetFgm( f: TFile; i: UFTP.TFgmIdx; VAR x: ARRAY OF BYTE ): CARDINAL;
VAR
  p: UFTP.TFileLen;
  l: CARDINAL;
BEGIN
  IF ( f = NIL ) OR ( i >= f^.n ) OR ( SIZE(x) < f^.l ) THEN RETURN 0 END;
  l := f^.l;
  p := i * l;
  IF f^.p # p THEN FIO.Seek( f^.f, p ) END;
  l := FIO.RdBin( f^.f, x, l );
  f^.p := FIO.GetPos( f^.f );
  RETURN l;
END GetFgm;

PROCEDURE Close( VAR f: TFile );
BEGIN
  IF f = NIL THEN RETURN END;
  FIO.Close( f^.f );
  Lib2.Fill( f^, 0 );
  f := NIL;
END Close;


PROCEDURE InitModule();
(*
VAR
  c: MD5.TCtxDesc;
*)
BEGIN
  Lib2.Fill( arr, 0 );
  Lib2.Fill( md5Null, 0 );
  (*
  MD5.Init( c );
  MD5.Final( c, md5Null );
  *)
END InitModule;

BEGIN
  InitModule();
END Files.
